from skopt import Optimizer as SKOptimizer
from skopt.utils import create_result
from skopt.plots import plot_convergence, plot_evaluations, plot_objective
from IPython.display import clear_output
import matplotlib.pyplot as plt

plt.style.use('dark_background')

class Optimizer:
    def __init__(self, model, featureset, target, validator, goal='maximize', search_spaces=None):
        if isinstance(search_spaces, type(None)):
            self.search_spaces = type(model).search_spaces
        else:
            self.search_spaces = search_spaces
        self.parameter_names = list(self.search_spaces.keys())
        self.value_spaces = list(self.search_spaces.values())
        self.opt = SKOptimizer(self.value_spaces)
        self.model = model
        self.featureset = featureset
        self.target = target
        self.validator = validator
        self.goal = goal
        if self.goal == 'maximize':
            self.coeff = -1
        elif self.goal == 'minimize':
            self.coeff = 1
        else:
            raise ValueError('Goal should be either to maximize or minimize objective.')
    
    def optimize(self, n_iters):
        for i in range(n_iters):
            pt = self.opt.ask()
            self.model.params = {k: v for k, v in zip(self.parameter_names, 
                                                      pt)}
            val = self.coeff * self.validator.score(self.model, self.featureset)
            clear_output(True)
            plot_convergence(self.opt.tell(pt, val));
            plt.show()

    def plot_objective(self):
        res = create_result(Xi=self.opt.Xi, 
                            yi=self.opt.yi, 
                            space=self.opt.space, 
                            rng=self.opt.rng, 
                            models=self.opt.models)
        plot_objective(res, dimensions=self.parameter_names);
        plt.show()
        
        
    def plot_evaluations(self):
        res = create_result(Xi=self.opt.Xi, 
                            yi=self.opt.yi, 
                            space=self.opt.space, 
                            rng=self.opt.rng, 
                            models=self.opt.models)
        plot_objective(res, dimensions=self.parameter_names);
        plt.show()


import numpy as np
from hashlib import sha256
import json

class Model:
    Estimator = None
    tracked_params = []
    fit_params = dict()
    short_name = 'model'

    def __init__(self, *args, **kwargs):
        self.estimator = self.Estimator(*args, **kwargs)
        self.params = {key: self.estimator.get_params()[key] for key in self.tracked_params}
        self.reset_name()
        self.is_fit = 0

    def reset_name(self):
        self.__name__ = f"{self.short_name}_{sha256((json.dumps(self.params, sort_keys=True)).encode()).hexdigest()[-2:] if self.params else 'default'}"

    def fit(self, X, y, **fit_params):
        if self.is_fit:
            raise UserWarning(
                f"This {type(self).__name__} is already fit.\nUse .warm_fit() to fit it from current state.")
        self.estimator.fit(X, y, **{key: fit_params[key]
                                    if key in fit_params
                                    else type(self).fit_params[key]
                                    for key in (set(fit_params.keys()) | set(type(self).fit_params.keys()))})
        self.is_fit = 1

    def predict(self, X, **predict_params):
        return self.estimator.predict(X, **predict_params)
    
    def __mul__(self, x):
        return WeightedModel(self, x)
    
    def __rmul__(self, x):
        return self * x
    
    def __truediv__(self, x):
        assert x != 0
        return WeightedModel(self, 1. / x)
    
    def __add__(self, other):
        # assert type(self).__base__ == type(other).__base__, \
        # f"Can't add {type(other).__name__} to {type(self).__name__}."
        if other == 0:
            return self
        return Ensemble([self, other])
    
    def __radd__(self, other):
        return self + other
        
    def __repr__(self):
        return self.__name__
        

class WeightedModel(Model): # MulNode
    """
    Multiplies predictions by a certain coefficient.
    """
    def __init__(self, model, coeff):
        self.model = model
        self.coeff = coeff
        if type(self.model) == type(self):
            self.coeff *= self.model.coeff
            self.model = self.model.model
        self.__name__ = f"{round(self.coeff, 2)} * "
        if '+' in self.model.__name__:
            self.__name__ += f"({self.model.__name__})"
        else:
            self.__name__ += f"{self.model.__name__}"

    def predict(self, X):
        """
        Standard prediction method.
        :param X: data matrix
        :return:
        """
        return self.coeff * self.model.predict(X)
    

class Ensemble(Model): # AddNode
    """
    Sums up all predictions of all models.
    """
    def __init__(self, models):
        self.models = []
        ensembles = [model for model in models if type(model) == type(self)]
        single_models = [model for model in models if type(model) != type(self)]
        _models = []
        for ensemble in ensembles:
            _models += ensemble.models
        _models += single_models

        def __get_model_coeff(model):
            """
            :param model:
            :return: if the model is weighted, then returns its coefficient, else 1
            """
            if isinstance(model, WeightedModel):
                return model.model, model.coeff
            return model, 1
        
        self.models = list(set([__get_model_coeff(model)[0] for model in _models]))
        _coeffs = np.zeros(len(self.models))
        for i in range(len(self.models)):
            for model in _models:
                if __get_model_coeff(self.models[i])[0] == __get_model_coeff(model)[0]:
                    _coeffs[i] += __get_model_coeff(model)[1]

        self.models = [WeightedModel(model, coeff) for model, coeff in zip(self.models, _coeffs)]
        self.norm_coeff = _coeffs.sum()
        _coeffs /= self.norm_coeff
        self.__name__ = ' + '.join([(model / self.norm_coeff).__name__ for model in self.models])
        
    def predict(self, X):
        """
        Standard prediction method.
        :param X: data matrix
        :return:
        """
        res = 0
        for model in self.models:
            res += model.predict(X)
        return res / self.norm_coeff
    
    def __mul__(self, x):
        return Ensemble([model * x for model in self.models])
    
    def __truediv__(self, x):
        assert x != 0
        return Ensemble([model / x for model in self.models])


class CustomModel(Model):
    tracked_params = []
    fit_params = dict()
    short_name = 'model'

    def __init__(self, model, tracked_params=[], fit_params=dict(), short_name=None):
        self.estimator = model
        self.short_name = (short_name if short_name is not None else self.get_short_name(model))
        self.tracked_params = tracked_params
        self.fit_params = fit_params
        self.params = {key: self.estimator.get_params()[key] for key in self.tracked_params}
        self.reset_name()
        self.is_fit = 0

    @staticmethod
    def get_short_name(instance):
        return instance.__class__.__name__.lower().replace('classifier', '_clf').replace('regressor', '_rg')


def model(model, tracked_params=[], short_name=None, fit_params={}):
    return CustomModel(model, tracked_params, fit_params, short_name)


import os


def check_structure(paths):
    for path in paths:
        if not os.path.isdir(path):
            return False
    return True


IMPORT_ERROR_MESSAGE = """
This directory doesn't look like kts project.
Use `kts init` to initialize a project. You can't use kts without its file system.
This error could also be raised by importing kts from a directory of existing project other than /notebooks.
"""


def check_file_system():
    paths = ['../input', '../notebooks', '../storage/info', '../storage/sources', '../output']

    if check_structure(paths):
        return

    raise ImportError(IMPORT_ERROR_MESSAGE)



from ..storage import source_utils
from ..storage.caching import cache
from .. import config
from .storage import FeatureConstructor
from ..storage import dataframe
from IPython.display import display
from glob import glob
import os


def preview(df, sizes=(2, 4, 6)):
    """
    Applies function to heads of particular dataframe.
    
    Example:
    ``` python
    @preview(df, sizes=[5, 15])
    def make_ohe_pclass(df):
        ...
    ```
    """

    def __preview(function):
        config.preview_call = 1
        try:
            for sz in sizes:
                # if isinstance(df, dataframe.DataFrame):
                #     ktdf = dataframe.DataFrame(df.head(sz), df.train, df.encoders)
                # else:
                #     ktdf = dataframe.DataFrame(df.head(sz), True, {})
                ktdf = dataframe.DataFrame(df.head(sz), True, {})
                display(function(ktdf))
        except Exception as e:
            config.preview_call = 0
            raise e
        config.preview_call = 0

    return __preview


def register(*args, cache_default=True):
    """
    Registers function for further caching its calls and restoring source.
    
    Example:
    ``` python
    @register
    def make_ohe_pclass(df):
        ...
    ```
    """

    def __register(func):
        # if source_utils.source_is_saved(func) and not source_utils.matches_cache(func):
        if func.__name__ + '_fc' in cache.cached_objs() and source_utils.get_source(func) != cache.load_obj(func.__name__ + '_fc').source:
            raise NameError("A function with the same name is already registered")

        if func.__name__ + '_fc' in cache.cached_objs():
            return cache.load_obj(func.__name__ + '_fc')
        else:
            functor = FeatureConstructor(func, cache_default)
            cache.cache_obj(functor, functor.__name__ + '_fc')
            return functor

    if args:
        function = args[0]
        return __register(function)
    else:
        return __register


def deregister(name, force=False):
    """
    Deletes sources and cached calls of a certain function.
    Usage:

    ``` python
    deregister('make_new_features')

    deregister('make_new_features', force=True)
    ```
    """
    confirmation = ''
    if not force:
        print("Are you sure you want to delete all the cached files?")
        print("To confirm please print full name of the function:")
        confirmation = input()
    if not force and confirmation != name:
        print("Doesn't match")
        return

    fc_name = name + '_fc'
    if fc_name in cache.cached_objs():
        print(f'removing {fc_name}')
        cache.remove_obj(fc_name)
    df_names = [df_name for df_name in cache.cached_dfs() if df_name.startswith(name + '__')]
    for df_name in df_names:
        print(f'removing {df_name}')
        cache.remove_df(df_name)


def dropper(function):
    """
    Registers function that won't be cached.
    Is recommended to be used only with functions which actually drop columns or rows and don't produce any new data.
    
    Example:
    ``` python
    @dropper
    def drop_pclass(df):
        return stl.column_dropper(['Pclass'])(df)
    ```
    """
    #     TODO:
    #     if cache.is_cached(function.__name__):
    #         print('Dropper is already registered. Deregistering: ')
    #         deregister(function.__name__, force=True)
    deregister(function.__name__, force=True)
    return register(function, cache_default=False)


def selector(function):
    """
    Registers function that won't be cached.
    Is recommended to be used only with functions which actually select columns or rows and don't produce any new data.
    
    Example:
    ``` python
    @selector
    def select_pclass_cabin(df):
        return stl.column_selector(['Pclass', 'Cabin'])(df)
    ```
    """

    deregister(function.__name__, force=True)
    return register(function, cache_default=False)


# def trainable(function):
#     def __trainable(df, )


import pandas as pd
from copy import deepcopy
import warnings


class SubDF(pd.DataFrame):
    def __getattr__(self, item):
        return None

    def __dict__(self):
        return dict()

class DataFrame(SubDF):
    """
    A wrapper over the standard DataFrame class.
    Complements it with .train and .encoders attributes.

    This class is implemented to supply an indicator
    for functions whether they serve a train or test call and let propagate
    this indicator further to inner functions.

    Returns kts.DataFrame from any attribute that produces pd.DataFrame,
    but kts.DataFrame.drop([]) will return pd.DataFrame.
    That's not a bug, as `a(b(df))` will still propagate
    aforementioned signal, but `a(b(df.drop()))` construction is not to be
    registered as a cached function.

    Example:
    ```
    def a(df):
        res = stl.empty_like(df)
        tmp = b(df)
        res['res'] = b['b'] ** 2
        return res
    ```
    """
    def __init__(self, df, train=None, encoders=None, slice_id=None):
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', UserWarning)

            if isinstance(df, DataFrame):
                super().__setattr__('df', df.df)
                super().__setattr__('slice_id', df.slice_id if isinstance(slice_id, type(None)) else slice_id)
                super().__setattr__('train', df.train if isinstance(train, type(None)) else train)  # ALERT: may cause errors during constructing like DF(DF(df), train=True)
                super().__setattr__('encoders', df.encoders if isinstance(encoders, type(None)) else encoders)  # not deepcopy to allow DF(df) init in FeatureConstructors
            else:
                super().__setattr__('df', df)
                super().__setattr__('slice_id', "0" * 16 if isinstance(slice_id, type(None)) else slice_id)
                super().__setattr__('train', False if isinstance(train, type(None)) else train)
                super().__setattr__('encoders', dict() if isinstance(encoders, type(None)) else encoders)

    def __dir__(self):
        return dir(self.df) + ['df', 'train', 'encoders', 'slice_id']

    def __copy__(self):
        return DataFrame(self.df, self.slice_id, self.train, deepcopy(self.encoders))

    def __getattr__(self, key):
        if key in ['train', 'encoders', 'df', 'slice_id']:
            return super().__getattr__(key)
        else:
            tmp = self.df.__getattr__(key)
            if isinstance(tmp, pd.DataFrame):
                return DataFrame(tmp, self.train, self.encoders)
            else:
                return tmp

    def __setattr__(self, key, value):
        if key in ['train', 'encoders', 'df', 'slice_id']:
            super().__setattr__(key, value)
        else:
            try:
                self.df.__setattr__(key, value)
            except:
                pass

    def __setitem__(self, key, value):
        self.df.__setitem__(key, value)

    def __getitem__(self, key):
        tmp = self.df.__getitem__(key)
        if isinstance(tmp, pd.DataFrame):
            return DataFrame(tmp, self.train, self.encoders)
        else:
            return tmp


def link(df, ktdf):
    assert isinstance(ktdf, DataFrame), 'Second dataframe should be of type KTDF, not pd.DF'
    return DataFrame(df, train=ktdf.train, encoders=ktdf.encoders, slice_id=ktdf.slice_id)


from . import classification
from . import binary_classification
from . import regression
from .. import modelling
import forge

exclude_types = [model_class for model_class in modelling.__dict__.values()
                 if isinstance(model_class, type) and modelling.Model in model_class.__bases__] + [modelling.Model]

model_classes = []
for module in (classification, binary_classification, regression):
    model_classes += [model_class for model_class in module.__dict__.values()
                      if isinstance(model_class, type) and modelling.Model in model_class.__bases__]

model_classes = set(model_classes) - set(exclude_types)

print(model_classes)

for model_class in model_classes:
    model_class.__init__ = forge.copy(model_class.Estimator.__init__)(modelling.Model.__init__)
    model_class.__doc__ = model_class.Estimator.__doc__




from ..modelling import *
from .. import config

from xgboost import XGBClassifier as _XGBC
class XGBClassifier(Model):
    Estimator = _XGBC
    default_params = {
        "max_depth": 3, 
        "learning_rate": 0.1, 
        "n_estimators": 100, 
        "booster": 'gbtree', 
        "gamma": 0, 
        "min_child_weight": 1, 
        "max_delta_step": 0, 
        "subsample": 1, 
        "colsample_bytree": 1, 
        "colsample_bylevel": 1, 
        "reg_alpha": 0, 
        "reg_lambda": 1, 
        "scale_pos_weight": 1, 
        "base_score": 0.5,
    }
    search_spaces = {
        'max_depth': (0, 50),
        'learning_rate': (0.01, 1.0, 'log-uniform'),
        'n_estimators': (50, 100),
        'booster': ['gbtree', 'dart'],
        'gamma': (1e-9, 0.5, 'log-uniform'),
        'min_child_weight': (0, 10),
        'max_delta_step': (0, 20),
        'subsample': (0.01, 1.0, 'uniform'),
        'colsample_bytree': (0.01, 1.0, 'uniform'),
        'colsample_bylevel': (0.01, 1.0, 'uniform'),
        'reg_alpha': (1e-9, 1.0, 'log-uniform'),
        'reg_lambda': (1e-9, 1000, 'log-uniform'),
        'scale_pos_weight': (1e-6, 500, 'log-uniform'),
        'base_score': (0.2, 0.7, 'uniform')
    }
    system_params = {
        "objective": 'binary:logistic', 
        "silent": True,
        "n_jobs": 1, 
        "random_state": 0, 
        "seed": config.seed
    }
    short_name = 'xgb'
    
    
from lightgbm import LGBMClassifier as _LGBMC
class LGBMClassifier(Model):
    Estimator = _LGBMC
    short_name = 'lgb'
    
    
from catboost import CatBoostClassifier as _CBC
class CBClassifier(Model):
    Estimator = _CBC
    short_name = 'cb'
    
    def predict(self, X):
        return self.estimator.predict_proba(X)
    